<?php

namespace Phem\Libraries\ImportExport\Xml\XmlSerialization;

use DOMDocument;
use Phem\Annotation;
use Phem\Core\Object;
use ReflectionClass;
use SimpleXMLElement;

/**
 * @author Jakub Petržílka <petrzilka@czweb.net>
 * 
 * Annotation driver XML serializer/unserializer
 * Only 4 annotations are used (XmlObject, XmlElement, XmlAttribute, XmlValue)
 * 
 * TODO: Enhance support of annotation arguments (introduce argument names etc.)
 * TODO: Caching reflected classes (use doctrine annotations?)
 * TODO: Implement unserialize method
 * 
 */
class XmlSerializer extends Object
{

    private $xmlVersion;
    private $encoding;
    private $numericKeyPrefix;

    public function __construct(
    $xmlVersion = '1.0', $encoding = 'utf-8', $numericKeyPrefix = 'entry_'
    )
    {
        $this->xmlVersion = $xmlVersion;
        $this->encoding = $encoding;
        $this->numericKeyPrefix = $numericKeyPrefix;
    }

    public function getXmlVersion()
    {
        return $this->xmlVersion;
    }

    public function setXmlVersion($xmlVersion)
    {
        $this->xmlVersion = $xmlVersion;
    }

    public function getEncoding()
    {
        return $this->encoding;
    }

    public function setEncoding($encoding)
    {
        $this->encoding = $encoding;
    }

    public function getNumericKeyPrefix()
    {
        return $this->numericKeyPrefix;
    }

    public function setNumericKeyPrefix($numericKeyPrefix)
    {
        $this->numericKeyPrefix = $numericKeyPrefix;
    }

    /**
     * 
     * @param object $object The input object to be serialized
     * @param boolean $formatOutput If set to true the output XML string
     * will be formated. Default true
     * @return string XML string
     */
    public function serialize($object, $formatOutput = true)
    {
        $xmlDoc = $this->serializeObject($object);

        $xmlDoc->formatOutput = $formatOutput;
        return $xmlDoc->saveXML();
    }

    private function serializeArray($xmlDoc, $root, $annotation, $fieldValue)
    {
        $dstObject = $root;
        $name = null;
        if (($annotation != null) && ($annotation->getParams() != null))
        {
            $name = $annotation->getParams()[0];
            if (array_key_exists(1, $annotation->getParams()) && (($annotation->getParams()[1] == "true") || ($annotation->getParams()[1] == 1)))
            {
                $dstObject = $xmlDoc->createElement($name);
                $root->appendChild($dstObject);
            }
        }

        foreach ($fieldValue as $key => $entry)
        {
            $finalName = $name;
            if ($name == null)
            {
                $finalName = $key;
            }

            if (is_int($finalName))
                $finalName = $this->getNumericKeyPrefix() . $finalName;

            if (is_array($entry))
            {
                $arrayElement = $xmlDoc->createElement($finalName);
                $dstObject->appendChild($arrayElement);
                $this->serializeArray($xmlDoc, $arrayElement, null, $entry);
            }
            else if (is_object($entry))
            {
                $dstObject->appendChild(
                        $this->serializeObject($entry, $xmlDoc, $finalName)
                );
            }
            else
            {
                $dstObject->appendChild(
                        $xmlDoc->createElement($finalName, $entry)
                );
            }
        }
    }

    private function serializeObject($object, $parent = null, $nodeName = null)
    {
        $className = get_class($object);
        $r = new ReflectionClass($className);

        $classAnnotations = Annotation::parse($r->getDocComment());

        /* check and validate class annotations */
        if (!($classAnnotations->containsKey('xmlobject')))
        {
            throw new XmlSerializerException(
            'XmlSerializer: Class '
            . get_class($object) .
            ' is missing XmlObject annotation'
            );
        }

        $disallowedAnnotations
                = array('xmlelement', 'xmlattribute', 'xmlvalue');

        if ($classAnnotations->containsAnyKeyOf($disallowedAnnotations))
        {
            throw new XmlSerializerException(
            'XmlSerializer: Annotations XmlElement,XmlAttribute and XmlValue are not allowed in class definition (' . get_class($object) . ') ');
        }

        /* create xml document and root element */

        /* for root class */
        if ($parent == null)
        {
            $xmlDoc = new DOMDocument($this->xmlVersion, $this->encoding);
            $nodeName = $r->getShortName();
        }
        /* for referenced classes */
        else
        {
            $xmlDoc = $parent;
        }

        /* Designate name for root element of object */
        if (count($classAnnotations->get('xmlobject')->getParams()) == 0)
        {
            $name = $nodeName;
        }
        else if (count($classAnnotations->get('xmlobject')->getParams()) == 1)
        {
            $name = $classAnnotations->get('xmlobject')->getParams()[0];
        }
        else
        {
            throw new XmlSerializerException(
            'XmlSerializer: XmlObject annotation on class'
            . get_class($object) .
            ' can have either 0 or 1 argument'
            );
        }

        $root = $xmlDoc->createElement($name);

        /* for root class */
        if ($parent == null)
        {
            $xmlDoc->appendChild($root);
        }

        /*
         * Load all object properties (fields)
         * into array $fields[$fieldName] => $fieldValue 
         */

        $properties = $r->getProperties();

        $fields = array();
        foreach ($properties as $prop)
        {
            $getMethod = "get" . ucFirst($prop->name);
            if (!method_exists($object, $getMethod))
            {
                throw new XmlSerializerException(
                "XmlSerializer: Missing get method for property "
                . $prop->name . " of class "
                . get_class($object)
                );
            }

            $fields[$prop->name] = $object->$getMethod();
        }

        /*
         * Iterate over fields 
         * and process them according to their annotations 
         */

        foreach ($fields as $fieldName => $fieldValue)
        {

            $annotations = Annotation::parse(
                            $r->getProperty($fieldName)->getDocComment()
            );

            foreach ($annotations as $annotation)
            {
                if ($annotation->getName() == "xmlobject")
                {
                    throw new XmlSerializerException(
                    'XmlSerializer: Annotation '
                    . $annotation->getName() .
                    ' not allowed on class fields ('
                    . get_class($object) . '->' . $fieldName . ') '
                    );
                }
                if ($annotation->getName() == "xmlelement")
                {

                    $elementName = $fieldName;
                    if ($annotation->getParams()[0] != null)
                        $elementName = $annotation->getParams()[0];

                    if (is_array($fieldValue))
                    {
                        $this->serializeArray(
                                $xmlDoc, $root, $annotation, $fieldValue
                        );
                    }
                    else if (is_object($fieldValue))
                    {
                        $root->appendChild(
                                $this->serializeObject(
                                        $fieldValue, $xmlDoc, $elementName
                                )
                        );
                    }
                    else if ($fieldValue != null)
                    {

                        $root->appendChild(
                                $xmlDoc->createElement(
                                        $elementName, $fieldValue
                                )
                        );
                    }
                }
                if ($annotation->getName() == "xmlattribute")
                {
                    if ($fieldValue == null)
                        continue;
                    if (is_object($fieldValue) || is_array($fieldValue))
                    {
                        throw new XmlSerializerException(
                        'XmlSerializer: XmlAttribute expects the field to be scalar, '
                        . gettype($fieldValue) .
                        ' given (' . get_class($object)
                        . '->' . $fieldName . ') '
                        );
                    }

                    $attrName = $fieldName;
                    if ($annotation->getParams()[0] != null)
                        $attrName = $annotation->getParams()[0];

                    $attr = $xmlDoc->createAttribute($attrName);
                    $attr->value = $fieldValue;
                    $root->appendChild($attr);
                }
                if ($annotation->getName() == "xmlvalue")
                {
                    if (is_object($fieldValue) || is_array($fieldValue))
                    {
                        throw new XmlSerializerException(
                        'XmlSerializer: XmlValue expects the field to be scalar, '
                        . gettype($fieldValue) .
                        ' given (' . get_class($object) .
                        '->' . $fieldName . ') '
                        );
                    }
                    $node = $xmlDoc->createTextNode($fieldValue);
                    $root->appendChild($node);
                }
            }
        }

        if ($parent == null)
        {
            return $xmlDoc;
        }
        else
        {
            return $root;
        }
    }

    function unserializeObject($src, $className,$path)
    {
        /* CHECK FOR CLASS ANNOTATIONS */
        $r = new ReflectionClass($className);

        $classAnnotations = Annotation::parse($r->getDocComment());

        /* check and validate class annotations */
        if (!($classAnnotations->containsKey('xmlobject')))
        {
            throw new XmlSerializerException(
            'XmlSerializer: Class '
            . $className .
            ' is missing XmlObject annotation'
            );
        }

        $disallowedAnnotations
                = array('xmlelement', 'xmlattribute', 'xmlvalue');

        if ($classAnnotations->containsAnyKeyOf($disallowedAnnotations))
        {
            throw new XmlSerializerException(
            'XmlSerializer: Annotations XmlElement,XmlAttribute and XmlValue are not allowed in class definition (' . $className . ') ');
        }
        /* END CHECK FOR CLASS ANNOTATIONS */

        $object = new $className();

        /*
         * Load all object properties (fields)
         * into array $fields[$fieldName] => $fieldValue 
         */

        $properties = $r->getProperties();

        foreach ($properties as $prop)
        {
            $setMethod = "set" . ucFirst($prop->name);
            if (!method_exists($object, $setMethod))
            {
                throw new XmlSerializerException(
                "XmlSerializer: Missing set method for property "
                . $prop->name . " of class "
                . $className
                );
            }

            $annotations = Annotation::parse(
                            $prop->getDocComment()
            );

            foreach ($annotations as $annotation)
            {
                if ($annotation->getName() == "xmlvalue")
                {
                    $path = substr ($path, 0,strlen($path)-1);
                    $object->$setMethod((string)$src->xpath("/".$path)[0]);
                }
                else if ($annotation->getName() == "xmlattribute")
                {
                    $object->$setMethod((string)$src->xpath("/".$path."@" . $prop->name)[0][$prop->name]);
                }
                else if ($annotation->getName() == "xmlelement")
                {
                    $propName = $prop->name;
                    if (isset($annotation->getParams()[0])&&($annotation->getParams()[0]!="{default}"))
                        $propName = $annotation->getParams()[0];
                    
                    $classPathArray = explode('\\', $className);
                    $namespace = str_replace("\\".end($classPathArray), "", $className);
                    $childClassName = $namespace.'\\'.$annotation->getParams()[2];
                    if (isset($annotation->getParams()[2]) && ($annotation->getParams()[2] != "array") && class_exists($childClassName))
                    {
                        $object->$setMethod($this->unserializeObject($src, $childClassName, $path.$propName."/"));
                    }
                    else if (isset($annotation->getParams()[2]) && ($annotation->getParams()[2] == "array"))
                    {
                        $entries = array();
                        if ($annotation->getParams()[1] == 1)
                            $path .= $propName;
                        else 
                            $path = substr ($path, 0,strlen($path)-1);
                        $xpathResult = $src->xpath($path);
                        
                        if (isset($xpathResult[0]))
                        {
                        $result = $xpathResult[0];
                        
                        
                        
                        //stupid detection of duplicate keys - TODO: remake
                        $first_key = "";
                        $duplicates = false;
                        foreach ($result as $key => $entry)
                        {
                            if ($first_key == "") $first_key = $key;
                            else
                            {
                                if ($first_key == $key)
                                {
                                    $duplicates = true;
                                    break;
                                }
                            }
                        }
                        
                        
                        $order = 1;
                        foreach ($result as $key => $entry)
                        {
                            $entryClassName = $namespace.'\\'.$annotation->getParams()[3];
                            if (isset($annotation->getParams()[3])&&($annotation->getParams()[3] != "array")&& class_exists($entryClassName))
                            {
                                $content = $this->unserializeObject($src, $entryClassName, $path."/*[".$order++."]/");
                                
                                if (!$duplicates)
                                {
                                    $entries[$key] = $content;
                                }
                                else
                                {
                                    $entries[] = $content;
                                }
                            }
                        }
                        $object->$setMethod($entries);
                        }
                    }
                    else
                    {   

                        $result = $src->xpath($path.$propName);
                	if (isset($result[0]))
                            $object->$setMethod((string)$result[0]);
                    }
                }
            }
        }



        return $object;
    }

    function unserialize($xml, $className = null)
    {
        $src = new SimpleXMLElement($xml);

        if ($className == null)
        {
            $className = $src->xpath("local-name(//)");
        }

        if (!class_exists($className))
        {
            throw new XmlSerializerException(
            "XmlSerializer: Class " . $className . " not found"
            );
        }
        
        $classPathArray = explode('\\', $className);
        
        return $this->unserializeObject($src, $className, "/".end($classPathArray)."/");
    }

}
