<?php

namespace Phem\Libraries\ResourceLoading;

use Phem\Core\Collection;
use Phem\Core\CollectionSortType;
use Phem\Core\Object;
use Phem\Environment\Resource;
use Phem\Environment\ResourceType;

/**
 * @author Jakub Petržílka <petrzilka@czweb.net>
 */
class ResourceManager extends Object
{

    private $resources;
    
    public function getResources()
    {
        return $this->resources;
    }    
    private function loadThirdPartyResources()
    {
        $resDir = FRAMEWORK_DIR
                .DS.BUILTIN_RESOURCES_DIR
                .DS.BUILTIN_THIRD_PARTY_RES_DIR;
        
        $dirContent = scandir($resDir);

        foreach ($dirContent as $dirEntry)
        {

            if (($dirEntry == ".") || ($dirEntry == ".."))
            {
                continue;
            }

            if (is_dir($resDir . DS . $dirEntry))
            {
                $loaderFileName = $resDir 
                        . DS . $dirEntry 
                        . DS . $dirEntry . RESOURCELOADER_SUFFIX . '.php';
                
                if (file_exists($loaderFileName))
                {
                    require_once $loaderFileName;

                    //the boot file must contain class named [SubdirectoryName]Boot
                    $loaderClass = $dirEntry . RESOURCELOADER_SUFFIX;
                    if (!class_exists($loaderClass))
                    {
                        die("Invalid loader file in third party resource " . $dirEntry);
                    }

                    //the class must be descendant of BootStrapper
                    $resourceLoader = new $loaderClass();
                    if (!is_a($resourceLoader, "Phem\Libraries\ResourceLoading\ThirdPartyResourceLoader"))
                    {
                        die("Loader of third part resource " . $dirEntry . " must be inherited from ThirdPartyResourceLoader");
                    }

                    $gainedResources = $resourceLoader->init();
                    foreach ($gainedResources as $gainedResource)
                    {
                        if (!($gainedResource instanceof Resource))
                        {
                            die("Third party resource " . $dirEntry . " returned invalid collection (all entries must be instance of Resouce)");
                        }
                    }
                    $this->resources->join($gainedResources);
                }
                else
                {
                    die("Loader file for third party resource " . $dirEntry . " not found");
                }
            }
        }
    }
    
    private function loadFrameworkResources()
    {
        //framework styles
        $cssRes = new Resource();
        
        $cssRes->setName("Framework styles");
        
        $cssRes->setPath(DS.FRAMEWORK_DIR
                .DS.BUILTIN_RESOURCES_DIR
                .DS.BUILTIN_CSS_DIR);
        
        $cssRes->setRecursiveLoad(true);
        $cssRes->setType(ResourceType::CSS);
        
        $this->resources->add($cssRes);
        
        //framework scripts
        $jsRes = new Resource();
        
        $jsRes->setName("Framework scripts");
        
        $jsRes->setPath(DS.FRAMEWORK_DIR
                .DS.BUILTIN_RESOURCES_DIR
                .DS.BUILTIN_JS_DIR);
        
        $jsRes->setRecursiveLoad(true);
        $jsRes->setType(ResourceType::JS);
        
        $this->resources->add($jsRes);        
    }
    
    private function loadApplicationResources()
    {
        //application styles
        $cssRes = new Resource();
        
        $cssRes->setName("Application styles");
        $cssRes->setPath(DS.CSS_DIR);
        $cssRes->setType(ResourceType::CSS);
        $cssRes->setRecursiveLoad(true);
        
        $this->resources->add($cssRes);
        
        //application scripts
        $jsRes = new Resource();
        
        $jsRes->setName("Application scripts");
        $jsRes->setPath(DS.JS_DIR);
        $jsRes->setType(ResourceType::JS);
        $jsRes->setRecursiveLoad(false);
        
        $this->resources->add($jsRes);           
    }    

    private function loadResources()
    {
        if ($this->resources != null) { return; }
        
        $this->resources = new Collection();
        
        $this->loadThirdPartyResources();
        
        $this->loadFrameworkResources();       
        $this->loadApplicationResources();       
        
        $this->resources->sort(
            CollectionSortType::NUMERIC
            |CollectionSortType::VALUE
            |CollectionSortType::DESC,
            "Priority"
        );

    }
    
    private function processDir($path,$recursive,$type,$initExtPath = null)
    {
        $pathCollection = new Collection();
        
        $resFilesRaw = scandir(getcwd().$path);

        foreach ($resFilesRaw as $resFile)
        {
            if (strpos($resFile, $type) > 0)
            {
                if ($initExtPath !== null)
                {
                    $ds = ($initExtPath === "")? "" : DS;
                    $pathCollection->add($initExtPath . $ds . $resFile);
                }
                else
                {
                    $pathCollection->add($path . DS . $resFile);
                }
            }
            else if (is_dir(getcwd().$path.DS.$resFile)&&($recursive)&&($resFile != ".")&&($resFile != ".."))
            {
                if ($initExtPath !== null)
                {
                    $ds = ($initExtPath === "")? "" : DS;
                    $pathCollection->join(
                        $this->processDir(
                            $path . DS . $resFile,$recursive,$type,$initExtPath.$ds.$resFile));                                 
                }
                else
                {
                    $pathCollection->join(
                        $this->processDir(
                            $path . DS . $resFile,$recursive,$type));             
                }                
            }
        }
        
        return $pathCollection;
    }
    
    public function getResourcePathCollection($type,$initExtPath = null)
    {
        $this->loadResources();
        
        $pathCollection = new Collection();
        
        foreach ($this->resources as $res)
        {
            
            /* @var $res Resource */
            if ($res->getType() != $type) { continue; }
            
            $fullPath = getcwd().$res->getPath();

            if (is_file($fullPath))
            {
                $pathCollection->add($res->getPath());
            }
            else if (is_dir($fullPath))
            {
                $pathCollection->join(
                    $this->processDir(
                        $res->getPath(),
                        $res->getRecursiveLoad(),
                        $res->getType(),$initExtPath));
            }
            else 
            {
                die($fullPath . " not found");
            }
        }
        
        return $pathCollection;
    }
    
    public function getMinifiedResourcePack($type)
    {
        $b = "";
        switch ($type)
        {
            default:
            case ResourceType::CSS:
                $b = "css";
                break;
            case ResourceType::JS:
                $b = "javascript";
                break;            
        }
        
        $paths = $this->getResourcePathCollection($type, "");
                
	$includePath= "";

        foreach ($paths as $file)
        {
            if ($includePath === "")
            {
                $includePath = "http://" . RESOURCES_SUBDOMAIN . "/minify/min/b=".$b."&f=".$file;
            }
            else
            {
                $includePath .= ",".$file;
            }
        }        
        
        return $includePath;
    }
}
        